// -----------------------------------------------------------------------------
//
// Copyright 2012(c) Analog Devices, Inc.
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//  - Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  - Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  - Neither the name of Analog Devices, Inc. nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//  - The use of this software may or may not infringe the patent rights
//    of one or more patent holders.  This license does not release you
//    from the requirement that you obtain separate licenses from these
//    patent holders to use this software.
//  - Use of the software either in source or binary form, must be run
//    on or directly connected to an Analog Devices Inc. component.
//
// THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// INTELLECTUAL PROPERTY RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// -----------------------------------------------------------------------------
// FILE NAME : ADC.v
// MODULE NAME : ADC
// AUTHOR : acostina
// AUTHORS EMAIL : adrian.costina@analog.com
// -----------------------------------------------------------------------------
// SVN REVISION: $WCREV$
// -----------------------------------------------------------------------------
// KEYWORDS : AD7490
// -----------------------------------------------------------------------------
// PURPOSE : Driver for the AD7490 16-Channel, 1 MSPS, 12-Bit ADC with
// Sequencer
// -----------------------------------------------------------------------------
// REUSE ISSUES
// Reset Strategy      : Active low reset signal
// Clock Domains       : fpga_clk_i 100 MHz input clock. The FPGA_CLOCK_FREQ 
//                       parameter must be adjusted if the input clock is changed.
//                       adc_clk_i is a 20 MHz clock
// Critical Timing     :
// Test Features       : N/A
// Asynchronous I/F    : N/A
// Instantiations      : N/A
// Synthesizable (y/n) : Y
// Target Device       : Cyclone IV E: EP4CE22F17C7 on BeMicroSDK kit
// Other               :
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

`timescale 1 ns / 100 ps //Use a timescale that is best for simulation.

//------------------------------------------------------------------------------
//----------- Module Declaration -----------------------------------------------
//------------------------------------------------------------------------------

module ADC
//----------- Ports Declarations -----------------------------------------------
(
    //clock and reset signals
    input               fpga_clk_i,     // system clock
    input               adc_clk_i,      // ADC clock
    input               reset_n_i,      // active low reset signal

    //IP control and data interface
    input               wr_data_n_i,    // active low signal to initiate a data write to the ADC
    input       [31:0]  data_i,         // data_i[15:0] is used for control register, data_i[31:16] is used for shadow register
    output reg  [15:0]  data_o,         // data read from the ADC
    output reg          data_rd_ready_o,// when set to high the data read from the ADC is available on the data_o bus
    output reg          data_wr_ready_o,// used to signal the status of an ADC write: 0 - write in progress, 1 - write complete

    //AD7490 control and data interface
    input               adc_sdo_i,      // ADC DOUT pin
    output reg          adc_sdi_o,      // ADC DIN pin
    output              adc_sclk_o,     // ADC SCLK pin
    output reg          adc_cs_n_o      // ADC \CS pin
);

//------------------------------------------------------------------------------
//----------- Registers Declarations -------------------------------------------
//------------------------------------------------------------------------------
reg [7:0]   adc_state;              // current state for the ADC control state machine
reg [7:0]   adc_next_state;         // next state for the ADC control state machine
reg [31:0]  data_i_reg ;            // register used on the fpga clock
reg [15:0]  data_adc;               // register used on the ADC clock
reg         adc_clk_en;             // clock enable signal
reg         sync_clk_d1;            // register used to synchronize ADC clock with fpga clock
reg         sync_clk_d2;            // register used to synchronize ADC clock with fpga clock
reg         sync_reg;               // register used to synchronize ADC clock with fpga clock
reg [8:0]   cycle_cnt;              // cycle time
reg [4:0]   sclk_cnt;               // count the sclk periods sent
reg [3:0]   delay_cs;               // registerused for increasing the time between CS Low and reading data
reg [3:0]   power_up;               // register used for the power_up sequence
reg         shadow_on;              // if 1, the shadow register must be written
reg         shadow_write;           // if 1 write shadow register, if 0 write control register

//------------------------------------------------------------------------------
//----------- Local Parameters -------------------------------------------------
//------------------------------------------------------------------------------
//ADC states
localparam ADC_IDLE_STATE           = 8'b00000001; // Default state
localparam ADC_START_CONV_STATE     = 8'b00000010; // Start the conversion
localparam ADC_RW_DATA_STATE        = 8'b00000100; // Generate clock signals for reading data from ADC
localparam ADC_DELAY_CS_STATE       = 8'b00001000; // Adds the t11 delay
localparam ADC_TRANSFER_DATA_STATE  = 8'b00010000; // Transfers data to the upper module
localparam ADC_WAIT_END_STATE       = 8'b00100000; // Waits for the cycle time to end. Includes the tQuiet timing

//ADC timing
localparam real FPGA_CLOCK_FREQ     = 100;  // FPGA clock frequency [MHz]
localparam real ADC_CYCLE_TIME      = 1;    // minimum time between two ADC conversions (Tcyc) [us]

localparam [8:0] ADC_CYCLE_CNT     = FPGA_CLOCK_FREQ * ADC_CYCLE_TIME - 2;

// ADC serial clock periods
localparam ADC_SCLK_PERIODS         = 5'd16  ; // number of clocks to be sent to the ADC to read the conversion result

//------------------------------------------------------------------------------
//------------------ Initial Block ---------------------------------------------
//------------------------------------------------------------------------------

initial
begin
    cycle_cnt       = ADC_CYCLE_CNT;
    sclk_cnt        = ADC_SCLK_PERIODS;
end

//------------------------------------------------------------------------------
//----------- Assign/Always Blocks ---------------------------------------------
//------------------------------------------------------------------------------

assign adc_sclk_o = adc_clk_en ? adc_clk_i : 1'b1;

// Clock synchronization process
// Makes sure that the clock enable signal is asserted during the high time of
// the ADC clock.
always @(posedge fpga_clk_i)
begin
    sync_clk_d1 <= adc_clk_i;
    sync_clk_d2 <= sync_clk_d1;
    sync_reg    <= sync_clk_d1 | ~sync_clk_d2;
end

// Process for the timers used by the module
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        cycle_cnt       <= ADC_CYCLE_CNT;
    end
    else
    begin
        //cycle count timer
        if(adc_state == ADC_IDLE_STATE)
        begin
            cycle_cnt   <= ADC_CYCLE_CNT;
        end
        else if (cycle_cnt != 0)
        begin
            cycle_cnt   <= cycle_cnt - 9'h1;
        end

        // timer used for respecting t2/t3b/t9 timings
        if (adc_state == ADC_IDLE_STATE)
        begin
            delay_cs    <= 4'h2;
        end
        else if (delay_cs != 4'h0)
        begin
            delay_cs <= delay_cs - 4'h1;
        end
    end
end

// Read Data from the ADC
always @(negedge adc_clk_i)
begin
    if(adc_clk_en == 1'b1)
    begin
        data_o          <= { data_o[14:0], adc_sdo_i };
        sclk_cnt        <= sclk_cnt - 5'h1;
    end
    else
    begin
        sclk_cnt        <= ADC_SCLK_PERIODS;
    end
end

// Write Data to the ADC
always @(posedge adc_clk_i)
begin
    if(adc_clk_en == 1'b1)
    begin
        if (power_up == 4'h0)
        begin
            adc_sdi_o   <= data_adc [15];
            data_adc    <= {data_adc[14:0],1'b0};
        end
        else
        begin
            adc_sdi_o   <= 1'b1;
        end
    end
    else
    begin
        if(shadow_write == 1'b1) // write the SHADOW register
        begin
            data_adc    <= {data_i_reg[30:16],1'b0} ;

            if (adc_state == ADC_START_CONV_STATE)
            begin
                adc_sdi_o   <= data_i_reg [31];
            end
            else
            begin
                adc_sdi_o   <= 1'b0;
            end
        end
        else                    // write the CONTROL register
        begin
            data_adc    <= {data_i_reg[14:0],1'b0 };

            if (adc_state == ADC_START_CONV_STATE)
            begin
                adc_sdi_o   <= data_i_reg [15];
            end
            else
            begin
                adc_sdi_o   <= 1'b0;
            end
        end
    end
end

// Update the ADC current state and the control signals
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        adc_state       <= ADC_IDLE_STATE;
        data_i_reg      <= 32'h8300;// write on, sequencer disabled, power normal,
                                    // tristate on, range 0 to 2xREFin , coding two's complement
        power_up        <= 4'h3;
        shadow_write    <= 1'b0;
    end
    else
    begin
        adc_state <= adc_next_state;
        case (adc_state)
            ADC_IDLE_STATE:
            begin

                if(power_up != 4'h0 )
                begin
                    power_up        <= power_up - 4'h1;
                end
                if( wr_data_n_i == 1'b0 && power_up == 4'h0 && shadow_write == 1'b0) //only read data after the powering up sequence
                begin
                    data_i_reg      <= data_i;
                    data_wr_ready_o <= 1'b1;
                    shadow_on       <= !data_i[14] & data_i[7]; // activate SHADOW register write if SEQ bit is 0 and SHADOW bit is 1
                    shadow_write    <= 1'b0;                    // the first write must be done to the CONTROL register
                end
                else
                begin
                    data_wr_ready_o <= 1'b0;
                end

                if (shadow_write == 1'b0 && shadow_on == 1'b1 ) // if the SHADOW register has been written once
                begin
                    data_i_reg      <= 32'h0;
                    shadow_write    <= 1'b0;
                    shadow_on       <= 1'b0;
                end

                adc_cs_n_o      <= 1'b1;
                data_rd_ready_o <= 1'b0;
                adc_clk_en      <= 1'b0;
            end
            //ADC write states
            ADC_START_CONV_STATE:
            begin
                adc_cs_n_o      <= 1'b0;
                data_wr_ready_o <= 1'b0;
                data_rd_ready_o <= 1'b0;
                adc_clk_en      <= 1'b0;
            end
            ADC_RW_DATA_STATE:
            begin
                adc_cs_n_o      <= 1'b0;
                data_wr_ready_o <= 1'b0;
                data_rd_ready_o <= 1'b0;
                adc_clk_en      <= 1'b1;
            end
            ADC_DELAY_CS_STATE:
            begin
                adc_cs_n_o      <= 1'b0;
                data_wr_ready_o <= 1'b0;
                data_rd_ready_o <= 1'b0;
                adc_clk_en      <= 1'b0;
            end
            ADC_TRANSFER_DATA_STATE:
            begin
                if (shadow_on == 1'b1)
                begin
                    shadow_write    <= !shadow_write; // write the SHADOW register the next iteration after writing the CONTROL register
                end
                adc_cs_n_o      <= 1'b1;
                data_wr_ready_o <= 1'b0;
                data_rd_ready_o <= 1'b1;
                adc_clk_en      <= 1'b0;
            end
            ADC_WAIT_END_STATE:
            begin
                adc_cs_n_o      <= 1'b1;
                data_wr_ready_o <= 1'b0;
                data_rd_ready_o <= 1'b0;
                adc_clk_en      <= 1'b0;
            end
        endcase
    end
end

//update the ADC next state
always @(adc_state, delay_cs, sclk_cnt, cycle_cnt, sync_reg )
begin
    adc_next_state = adc_state;
    case (adc_state)
    //ADC IDLE state
        ADC_IDLE_STATE:
        begin
            adc_next_state = ADC_START_CONV_STATE;
        end
    //ADC write states
        ADC_START_CONV_STATE:
        begin
            if (delay_cs == 4'h0 && sync_reg == 1'b0 )
            begin
                adc_next_state = ADC_RW_DATA_STATE;
            end
        end
        ADC_RW_DATA_STATE:
        begin
            if(sclk_cnt == 5'h0)
            begin
                adc_next_state = ADC_DELAY_CS_STATE;
            end
        end
        ADC_DELAY_CS_STATE:
        begin
            adc_next_state = ADC_TRANSFER_DATA_STATE;
        end
        ADC_TRANSFER_DATA_STATE:
        begin
            adc_next_state = ADC_WAIT_END_STATE;
        end
        ADC_WAIT_END_STATE:
        begin
            if ( cycle_cnt == 9'b0 )
            begin
                adc_next_state = ADC_IDLE_STATE;
            end
        end
        // default action
        default:
        begin
            adc_next_state = ADC_IDLE_STATE;
        end
    endcase
end

endmodule
